/**
 * Copyright (C) Mellanox Technologies Ltd. 2015-2016.  ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */

#ifndef _SHARP_COLL_API_H
#define _SHARP_COLL_API_H
#include <stdint.h>


/**
 * @brief SHArP coll supported data types
 *
 * The enumeration list describes the datatypes supported by SHArP coll
 */
#ifdef __cplusplus
extern "C" {
#endif

enum sharp_datatype {
	SHARP_DTYPE_UNSIGNED,		/**< 32-bit unsigned integer. */
	SHARP_DTYPE_INT,		/**< 32-bit integer. */
	SHARP_DTYPE_UNSIGNED_LONG,	/**< 64-bit unsigned integer. */
	SHARP_DTYPE_LONG,		/**< 64-bit integer. */
	SHARP_DTYPE_FLOAT,		/**< 32-bit long floating point number */
	SHARP_DTYPE_DOUBLE,		/**< 64-bit long floating point number. */
	SHARP_DTYPE_NULL		/**< NULL data type */
};

/**
 * @brief SHArP coll supported aggregation operations
 *
 * The enumeration list describes the aggregation operations supported by SHArP coll
 */
enum sharp_reduce_op {
	SHARP_OP_MAX,		/**< maximum. */
	SHARP_OP_MIN,		/**< minimum. */
	SHARP_OP_SUM,		/**< sum. */
	SHARP_OP_PROD,		/**< product. */
	SHARP_OP_LAND,		/**< logical and.*/
	SHARP_OP_BAND,		/**< bit-wise and. */
	SHARP_OP_LOR,		/**< logical or. */
	SHARP_OP_BOR,		/**< bit-wise or. */
	SHARP_OP_LXOR,		/**< logical xor. */
	SHARP_OP_BXOR,		/**< bit-wise xor. */
	SHARP_OP_MAXLOC,	/**< max value and location. */
	SHARP_OP_MINLOC,	/**< min value and location. */
	SHARP_OP_NULL
};

/**
 * @brief SHArP coll status code
 *
 * The enumeration list describes the error codes returned by SHArP coll
 */
enum sharp_error_no {
	SHARP_COLL_SUCCESS	= 0,	/**< Success. */
	SHARP_COLL_ERROR	= -1,	/**< Error. */
	SHARP_COLL_ENOT_SUPP	= -2,	/**< Collective operation not supported. */
	SHARP_COLL_ENOMEM	= -3,	/**< No memory. */
	SHARP_COLL_EGROUP_ALLOC	= -4,	/**< SHArP Group alloc error. */
	SHARP_COLL_ECONN_TREE	= -5,	/**< No connection to sharp tree. */
	SHARP_COLL_EGROUP_JOIN	= -6,	/**< Not able to join sharp grou.*/
	SHARP_COLL_EQUOTA	= -7,	/**< SHArP resource quota error. */
	SHARP_COLL_ESESS_INIT	= -8,	/**< Cannot connect to SHArPD. */
	SHARP_COLL_EDEV		= -9,	/**< SHArP device error. */
	SHARP_COLL_EINVAL	= -10,	/**< Invalid value. */
	SHARP_COLL_EJOB_CREATE	= -11,	/**< Cannot create SHArP job. */
	SHARP_COLL_ETREE_INFO	= -12,	/**< SHArP tree info not found. */
	SHARP_COLL_ENOTREE	= -13,	/**< No available SHArP trees. */
	SHARP_COLL_EGROUP_ID	= -14,	/**< Wrong SHArP group ID. */
	SHARP_COLL_EOOB		= -15,	/**< Out-Of-Band collective error. */
	SHARP_COLL_EGROUP_MCAST = -16,  /**< Multicast target error. */
	SHARP_COLL_EGROUP_TRIM	= -17  /**< SHArP Group trim failed. */
};

/* Forward declarations */
struct sharp_coll_context;
struct sharp_coll_comm;
struct sharp_coll_request;

/*
 * @brief SHArP coll configuration descriptor
 *
 * This descriptor defines the configuration for SHArP coll initialization
 */
struct sharp_coll_config {
	char *ib_dev_list;		/**< IB device name, port list.  */
	int user_progress_num_polls;	/**< Number of polls to do before calling user progress. */
	int coll_timeout;		/**< Timeout (msec) for collective operation, -1 - infinite */
};

/**
 * Default SHARP COLL configuration.
 */
extern const struct sharp_coll_config sharp_coll_default_config;

/*
 * @brief SHArP coll Out-Of-Band collectives descriptor
 *
 * This descriptor defines list of OOB collectives application must implement
 * and provide for SHArP coll initialization
 */
struct sharp_coll_out_of_band_colls {
	/**
	 * @brief out-of-band broadcast
	 *
	 * The pointer refers to application defined out-of-band bcast collective.
	 *
	 * @param [in] context	User-defined context or NULL.
	 * @param [in] buffer	Buffer to send/recv.
	 * @param [in] len	Size of the buffer.
	 * @param [in] root	Root of the the broadcast.
	 */
	int (*bcast) (void* context, void* buffer, int len, int root);

	/**
	 * @brief out-of-band barrier
	 *
	 * The pointer refers to application defined out-of-band barrier collective.
	 *
	 * @param [in] context	User-defined context or NULL.
	 */
	int (*barrier) (void* context);

	/**
	 * @brief out-of-band gather
	 *
	 * The pointer refers to application defined out-of-band gather collective.
	 *
	 * @param [in] context	User-defined context or NULL.
	 * @param [in] root	Root of the the broadcast.
	 * @param [in] sbuf	Buffer to send.
	 * @param [in] rbuf	Buffer to recv.
	 * @param [in] len	Size of the buffer.
	 */
	int (*gather) (void * context, int root, void *sbuf, void *rbuf, int len);
};

/*
 * @brief SHArP coll group initialization spec descriptor
 *
 * This descriptor defines the list of application specification to create SHArP group.
 *
 */
struct sharp_coll_comm_init_spec {
	int	rank;		/**< Uniq process rank in the group. */
	int	size;		/**< Size of the SHArP group. */
	int	is_comm_world;	/**< Is universal group (MPI_COMM_WORLD). */
	void	*oob_ctx;	/**< External group context for OOB functions. */
};

/*
 * @brief SHArP coll initialization spec descriptor
 *
 * This descriptor defines the list of application specification to initialize SHArP coll.
 *
 */
struct sharp_coll_init_spec {
	uint64_t	job_id;				/**< Job unique ID */
	char		*hostlist;			/**< Comma separated hostlist. */
	int		world_rank;			/**< Global unique process number. */
	int		world_size;			/**< Num of processes in the job. */
	int		(*progress_func)(void);		/**< External progress function. */
	int		group_channel_idx;		/**< local group channel index(0 .. (max - 1))*/
	struct sharp_coll_config config;		/**< @ref sharp_coll_config "SHARP COLL Configuration". */
	struct sharp_coll_out_of_band_colls oob_colls;  /**< @ref sharp_coll_out_of_band_colls "List of OOB collectives". */
};


/**
 * @brief SHArP coll buffer types
 *
 * The enumeration list describes the buffer types supported in collective calls.
 * @note Only SHARP_DATA_BUFFER is implemented.
 */
enum sharp_data_buffer_type {
	SHARP_DATA_BUFFER,	/**< Contiguous buffer. */
	SHARP_DATA_STREAM,	/**< Data stream. */
	SHARP_DATA_IOV		/**< Vector input. */
};

/**
 * @brief stream callback
 *
 * Streaming writer callback function.
 *
 * @param buffer    Buffer to read or write the data
 * @param length    Max. number of bytes to copy.
 * @param offset    Offset in the data stream. First time called with offset = 0,
 *                  and it's incremented by the return value.
 * @param context   User-define context, from stream.cb_context;
 *
 * @return Number of bytes processed from the buffer.
 *         For writers - if it's smaller than `length', it means to end the
 *          fragment prematurely, and the callback will be called again until
 *          total length is processed.
 *         For readers, the callback must process the entire buffer, if it returns
 *          a smaller value it means the message is truncated.
 *
 * @note It's guaranteed that (offset + length) is not larger than the declared
 *       stream size (stream.length).
 *
 */
typedef size_t (*sharp_stream_cb)(void *buffer, size_t length,
				    size_t offset, void *context);

/*
 * @brief SHArP coll input buffer description
 *
 * This descriptor defines the buffer description for SHArP coll operations
 *
 */
struct sharp_coll_data_desc {
	enum sharp_data_buffer_type type;
	union {

		/* contiguous buffer */
		struct {
			void *ptr;		/**< contiguous data buffer. */
			size_t length;		/**< Buffer len. */
			void *mem_handle;	/**< memory handle returned from @ref sharp_coll_reg_mr */
		} buffer;

		/* Streaming data */
		struct {
			sharp_stream_cb cb;	/**< Streaming callback. */
			void *cb_context;	/**< Context for stream callback. */
			size_t length;		/**< Total data length. */
		} stream;

		/* Scatter/gather list */
		struct {
			unsigned count;		/**< Number of IOV entries. */
			struct iovec *iov;	/**< IOV entries. */
		} iov;

	};
};

/*
 * @brief SHArP coll reduce collective specification
 *
 * This descriptor defines the input paramaters for SHArP coll reduce operation
 *
 */
struct sharp_coll_reduce_spec {
	int				root;		/**< [in] root rank number (ignored for allreduce) */
	struct sharp_coll_data_desc	sbuf_desc;	/**< [in] source data buffer desc */
	struct sharp_coll_data_desc	rbuf_desc;	/**< [out] destination data buffer desc */
	enum sharp_datatype		dtype;		/**< [in] data type @ref sharp_datatype */
	int				length;		/**< [in] reduce operation size */
	enum sharp_reduce_op		op;		/**< [in] reduce operator @ref sharp_reduce_op */
};

/*
 * @brief SHArP coll broadcast collective specification
 *
 * This descriptor defines the input paramaters for SHArP coll broadcast operation
 *
 */
struct sharp_coll_bcast_spec {
	int				root;		/**< [in] root rank number */
	struct sharp_coll_data_desc	buf_desc;	/**< [in,out] buffer desc to send/recv bcast data */
	int				size;		/**< [in] bcast size */
};

/**
 * @brief SHArP coll context initialization
 *
 * This routine is initialize SHArP coll library and create @ref sharp_coll_context "SHARP coll context".
 * This is a collective, called from all processes of the job.
 *
 * @warning An application cannot call any SHArP coll routine before sharp_coll_init
 *
 * @param [in]	sharp_coll_spec		SHArP coll specification descriptor.
 * @param [out]	sharp_coll_context	Initialized @ref sharp_coll_context "SHArP coll context".
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_init(struct sharp_coll_init_spec *sharp_coll_spec,
		    struct sharp_coll_context  **sharp_coll_context);

/**
 * @brief SHArP coll context finalize
 *
 * This routine finalizes and releases the resources associated with
 * @ref sharp_coll_context "SHArP coll context". typically done once, just before the process ends.
 *
 * @warning An application cannot call any SHArP coll routine after sharp_coll_finalize
 *
 * @param [in] context	SHArP coll context to cleanup.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_finalize(struct sharp_coll_context *context);

/**
 * @brief Progress SHArP coll communication operations.
 *
 * This routine explicitly progresses all SHArP communication operation.
 * This typically called from MPI progress context for MPI case.
 *
 * @param [in] context	SHArP coll context to progress.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_progress(struct sharp_coll_context *context);

/**
 * @brief SHArP coll communicator(group) initialization
 *
 * This routine creates @ref sharp_coll_comm "SHArP coll group".
 * This is a collective, called from all processes of the SHArP group.
 *
 * @param [in]	context		Handle to SHArP coll context.
 * @param [in]	spec		Input @ref sharp_coll_comm_init_spec "SHArP coll group specification".
 * @param [out] sharp_coll_comm Handle to SHArP coll communicator(group)
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_comm_init(struct sharp_coll_context *context,
			 struct sharp_coll_comm_init_spec *spec,
			 struct sharp_coll_comm **sharp_coll_comm);

/**
 * @brief SHArP coll communicator cleanup
 *
 * This routine cleanup SHArP coll communicator handle returned from @ref sharp_coll_comm_init.
 *
 * @param [in] comm   SHArP coll cCommunicator to destroy.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_comm_destroy(struct sharp_coll_comm *comm);

/**
 * @brief SHArP coll barrier collective
 *
 * This routine is collective operation blocks until all processes call this routine .
 *
 * @param [in]	comm	SHArP coll communicator to run the barrier.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_do_barrier(struct sharp_coll_comm *comm);

/**
 * @brief SHArP coll non-blocking barrier collective
 *
 * This routine is non blocking version of @ref sharp_coll_do_barrier "SHArP coll barrier".
 * The progress of this operation is tracked with return request handle
 *
 * @param [in]	comm	SHArP coll communicator to run the barrier.
 * @param [out] handle  Handle representing the communication operation.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_barrier_nb(struct sharp_coll_comm *comm,
			     struct sharp_coll_request **handle);

/**
 * @brief SHArP coll allreduce collective
 *
 * This routine aggreate the data from all processes of the groupand
 * distributes the result back to all processes.
 *
 * @param [in]	comm	SHArP coll communicator to run the allreduce collective.
 * @param [in]	spec	Allreduce operation specification.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_do_allreduce(struct sharp_coll_comm *comm,
			     struct sharp_coll_reduce_spec *spec);

/**
 * @brief SHArP coll non-blocking allreduce collective
 *
 * This routine is non blocking version of @ref sharp_coll_do_allreduce "SHArP coll allreduce".
 * The progress of this operation is tracked with return request handle
 *
 * @param [in]	comm	SHArP coll communicator to run the allreduce collective.
 * @param [in]	spec	Allreduce operation specification.
 * @param [out] handle  Handle representing the communication operation.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_allreduce_nb(struct sharp_coll_comm *comm,
				struct sharp_coll_reduce_spec *spec,
				struct sharp_coll_request **req);

/**
 * @brief SHArP coll reduce collective
 *
 * This routine aggreate the data from all processes of the group to a specific process
 *
 * @param [in]	comm	SHArP coll communicator to run the reduce collective.
 * @param [in]	spec	Reduce operation specification.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_reduce(struct sharp_coll_comm *comm,
			 struct sharp_coll_reduce_spec *spec);

/**
 * @brief SHArP coll non-blocking reduce collective
 *
 * This routine is non blocking version of @ref sharp_coll_do_reduce "SHArP coll Reduce".
 * The progress of this operation is tracked with return request handle
 *
 * @param [in]	comm	SHArP coll communicator to run the reduce collective.
 * @param [in]	spec	Allreduce operation specification.
 * @param [out] handle  Handle representing the communication operation.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_reduce_nb(struct sharp_coll_comm *comm,
			    struct sharp_coll_reduce_spec *spec,
			    struct sharp_coll_request **handle);
/**
 * @brief SHArP coll broadcast collective
 *
 * This routine broadcast data from single process to all processes of the group
 *
 * @param [in]	comm	SHArP coll communicator to run the bcast collective.
 * @param [in]	spec	Bcast operation specification.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_bcast(struct sharp_coll_comm *comm,
			struct sharp_coll_bcast_spec *spec);

/**
 * @brief SHArP coll non blocking broadcast collective
 *
 * This routine is non-blocking version of @ref sharp_coll_do_bcast "SHArP coll Bcast".
 * The progress of this operation is tracked with return request handle
 *
 * @param [in]	comm	SHArP coll communicator to run the bcast collective.
 * @param [in]	spec	Bcast operation specification.
 * @param [out] handle  Handle representing the communication operation.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Not implemented.
 */
int sharp_coll_do_bcast_nb(struct sharp_coll_comm *comm,
			   struct sharp_coll_bcast_spec *spec,
			   struct sharp_coll_request **handle);

/**
 * @brief SHArP coll request test
 *
 * This routine tests for the completion of a specific non-blocking coll operation
 *
 * @param [in]	req	SHArP coll request handle
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_req_test(struct sharp_coll_request *req);

/**
 * @brief SHArP coll request wait
 *
 * This routine returns when the operation identified by non-blocking collective request is complete.
 * The request object is deallocated by the call to sharp_coll_req_wait and the request handle is set to NULL.
 *
 * @param [in]	req	SHArP coll request handle
 *
 * @return Error code as defined by @ref sharp_error_no
 */

int sharp_coll_req_wait(struct sharp_coll_request *req);

/**
 * @brief SHArP coll request deallocate
 *
 * This routine deallocates request handle
 *
 * @param [in]	req	SHArP coll request handle
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_req_free(struct sharp_coll_request *req);

/**
 * @brief SHArP coll translate MPI OP to SHArP coll OP
 *
 * This routine returns SHArP coll reduce operation ID for given MPI OP string if it is supported.
 *
 * @param [in]	mpi_op_str	MPI Operation name.
 *
 * @return SHArP reduce Op ID as defined by @ref sharp_reduce_op on success, < 0 on failure
 */
int sharp_translate_mpi_op(const char *mpi_op_str);

/**
 * @brief SHArP coll translate MPI datatype string to SHArP coll data type
 *
 * This routine returns SHArP coll data type for given MPI data type string if it is supported.
 *
 * @param [in] mpi_dtype_str  MPI Datatype string.
 *
 * @return: SHArP coll datatype as defined by @ref sharp_datatype  on success, < 0 on failure.
 */
int sharp_translate_mpi_dtype(const char *mpi_dtype_str);

/**
 * @brief SHArP coll memory registration.
 *
 * This routine registers external mem buffer
 *
 * @param [in] context	SHArP coll context to progress.
 * @param [in] buf	Buffer to register
 * @param [in] size	length of the buffer in bytes
 * @param [out] mr	memory registration handle.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note Only one outstanding registration supported. no registration cache.
 *
 */
int sharp_coll_reg_mr(struct sharp_coll_context *context,
		      void *buf, int size, void **mr);

/**
 * @brief SHArP coll memory de-registration.
 *
 * This routine de-registers the MR.
 *
 * @param [in] context	SHArP coll context to progress.
 * @param [in] mr	memory registration handle.
 *
 * @return Error code as defined by @ref sharp_error_no
 */
int sharp_coll_dereg_mr(struct sharp_coll_context *context, void *mr);

/**
 * @brief SHArP coll print config flags
 *
 * The enumeration list describes bit masks for different options to print config flags
 */
enum config_print_flags {
	SHARP_COLL_CONFIG_PRINT_CONFIG	= 1,	/**< basic configuration. */
	SHARP_COLL_CONFIG_PRINT_HEADER	= 2,	/**< Print header. */
	SHARP_COLL_CONFIG_PRINT_DOC	= 4,	/**< full description. */
	SHARP_COLL_CONFIG_PRINT_HIDDEN  = 8	/**< hidden options. */
};

/**
 * @brief SHArP coll print configuration
 *
 * This routine prints SHArP coll configuration to a stream.
 *
 * @param [in] stream		Output stream to print to.
 * @param [in] print_flags	Controls how the configuration is printed.
 *
 * @return Error code as defined by @ref sharp_error_no
 */

int sharp_coll_print_config(FILE *stream, enum config_print_flags print_flags);

/**
 * @brief SHArP coll print error string
 *
 * This routine returns error string for a given @ref sharp_error_no "SHArP coll error code".
 *
 * @param [in] error	SHArP coll rror code.
 *
 * @return Error string
 */
const char *sharp_coll_strerror(int error);

/**
 * @brief SHArP coll print statistics
 *
 * This routine dumps SHArP coll usage statistics
 *
 * @param [in] context	SHArP coll context to progress.
 *
 * @return Error code as defined by @ref sharp_error_no
 *
 * @note It is expected to Out-Of_Band collectives are operational valid to get
 *	 accumulated stats dumps (SHARP_COLL_STATS_DUMP_MODE=2) during the finalize process.
 *
 */
int sharp_coll_dump_stats(struct sharp_coll_context *context);

#ifdef __cplusplus
}
#endif

#endif
